import json
import os
import re


def snake_case(string):
    # From https://gist.github.com/jaytaylor/3660565#gistcomment-2271689
    return re.compile(r"(?!^)(?<!_)([A-Z])").sub(r"_\1", string).lower().replace(" ", "_").replace("-", "_")


def get_json_data(filename):
    with open(os.path.join(os.path.dirname(__file__), "data", filename), "r") as f:
        return json.load(f)


def parameter_transform(params):
    return {param["name"]: param for param in params}



import os
from collections import ChainMap
from collections import namedtuple

import pytest
import requests
from octokit import Octokit


class TestAuth(object):
    def test_can_set_basic_authentication(self):
        sut = Octokit(auth="basic", username="myuser", password="mypassword")
        assert sut.auth == "basic"
        assert sut.username == "myuser"
        assert sut.password == "mypassword"

    def test_cannot_set_basic_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth="basic")
        with pytest.raises(AssertionError):
            Octokit(auth="basic", username="z", password="")
        with pytest.raises(KeyError):
            Octokit(auth="basic", password="mypassword")
        with pytest.raises(AssertionError):
            Octokit(auth="basic", username="")
        with pytest.raises(KeyError):
            Octokit(auth="basic", username="xyz")

    def test_basic_auth_used_if_set(self, mocker):
        mocker.patch("requests.get")
        Octokit(auth="basic", username="myuser", password="mypassword").oauth_authorizations.get_authorization(
            authorization_id=100
        )
        requests.get.assert_called_once_with(
            "https://api.github.com/authorizations/100",
            params={},
            headers=Octokit().headers,
            auth=("myuser", "mypassword"),
        )

    def test_can_set_token_authentication(self):
        sut = Octokit(auth="token", token="yippy")
        assert sut.auth == "token"
        assert sut.token == "yippy"

    def test_cannot_set_token_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth="token")
        with pytest.raises(AssertionError):
            Octokit(auth="token", token="")

    def test_token_auth_used_if_set(self, mocker):
        mocker.patch("requests.get")
        Octokit(auth="token", token="yak").oauth_authorizations.get_authorization(authorization_id=100)
        headers = dict(ChainMap(Octokit().headers, {"Authorization": "token yak"}))
        requests.get.assert_called_once_with("https://api.github.com/authorizations/100", params={}, headers=headers)

    def test_can_set_installation_authentication(self, mocker):
        Request = namedtuple("Request", ["json"])
        get = mocker.patch("requests.get")
        get.return_value = Request(json=lambda: [{"id": 13, "app_id": 1}, {"id": 37, "app_id": 42}])
        post = mocker.patch("requests.post")
        post.return_value = Request(json=lambda: {"token": "v1.1f699f1069f60", "expires_at": "2016-07-11T22:14:10Z"})
        with open(os.path.join(os.path.dirname(__file__), "test.pem"), "r") as f:
            private_key = f.read()
        sut = Octokit(auth="installation", app_id="42", private_key=private_key)
        assert sut.auth == "installation"
        assert sut.token == "v1.1f699f1069f60"
        assert sut.expires_at == "2016-07-11T22:14:10Z"

    def test_cannot_set_installation_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth="installation")
        with pytest.raises(AssertionError):
            Octokit(auth="installation", app_id="")
        with pytest.raises(KeyError):
            Octokit(auth="installation", app_id="42")
        with pytest.raises(AssertionError):
            Octokit(auth="installation", app_id="42", private_key="")

    def test_installation_token_is_used_if_set(self, mocker):
        Request = namedtuple("Request", ["json"])
        get = mocker.patch("requests.get")
        get.return_value = Request(json=lambda: [{"id": 13, "app_id": 1}, {"id": 37, "app_id": 42}])
        post = mocker.patch("requests.post")
        post.return_value = Request(json=lambda: {"token": "v1.1f699f1069f60", "expires_at": "2016-07-11T22:14:10Z"})
        with open(os.path.join(os.path.dirname(__file__), "test.pem"), "r") as f:
            private_key = f.read()
        sut = Octokit(auth="installation", app_id="42", private_key=private_key)
        assert sut.installation_id == 37
        get = mocker.patch("requests.get")
        sut.oauth_authorizations.get_authorization(authorization_id=100)
        headers = {
            "Content-Type": "application/json",
            "Authorization": "token v1.1f699f1069f60",
            "accept": "application/vnd.github.machine-man-preview+json",
        }
        requests.get.assert_called_once_with("https://api.github.com/authorizations/100", params={}, headers=headers)

    def test_cannot_set_app_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth="app")
        with pytest.raises(AssertionError):
            Octokit(auth="app", app_id="")
        with pytest.raises(KeyError):
            Octokit(auth="app", app_id=42)
        with pytest.raises(AssertionError):
            Octokit(auth="app", app_id=42, private_key="")

    def test_can_set_app_authentication(self, mocker):
        Request = namedtuple("Request", ["json"])
        get = mocker.patch("requests.get")
        get.return_value = Request(json=lambda: [{"id": 37}])
        with open(os.path.join(os.path.dirname(__file__), "test.pem"), "r") as f:
            private_key = f.read()
        sut = Octokit(auth="app", app_id=42, private_key=private_key)
        assert sut.auth == "app"
        assert sut.jwt is not None

    def test_can_get_with_app_authentication(self, mocker):
        Request = namedtuple("Request", ["json"])
        get = mocker.patch("requests.get")
        get.return_value = Request(json=lambda: [{"id": 37}])
        with open(os.path.join(os.path.dirname(__file__), "test.pem"), "r") as f:
            private_key = f.read()
        sut = Octokit(auth="app", app_id=42, private_key=private_key)
        assert sut.apps.get_authenticated()

    def test_can_make_unauthenticated_call(self, mocker):
        mocker.patch("requests.get")
        Octokit().users.list_followers_for_user(username="octokit")
        requests.get.assert_called_once_with(
            "https://api.github.com/users/octokit/followers",
            headers={"Content-Type": "application/json", "accept": "application/vnd.github.v3+json"},
            params={"page": 1, "per_page": 30},
        )



import datetime
import json
import re
from collections import ChainMap
from collections import defaultdict

import requests
from jose import jwt
from octokit import errors
from octokit import utils


class Base(object):

    base_url = "https://api.github.com"

    def __init__(self):
        self.headers = {"accept": "application/vnd.github.v3+json", "Content-Type": "application/json"}
        self._attribute_cache = defaultdict(dict)

    def _get_headers(self, method_headers):
        return dict(ChainMap(method_headers, self.headers))

    def _get_required_params(self, params, cached_kwargs):
        all_required = [k for k, v in params.items() if v.get("required")]
        required_params = all_required
        return required_params

    def validate(self, parameters, definition):
        self.validate_required_parameters(parameters, self.get_required_parameters(definition))
        schema = {}
        if definition.get("requestBody"):
            schema = definition["requestBody"]["content"]["application/json"]["schema"]
            self.validate_schema(parameters, schema)
        properties = schema.get("properties", {})
        valid_parameters = [p["name"] for p in definition.get("parameters")] + list(properties.keys())
        self.validate_other_parameters(parameters, valid_parameters, properties)
        return True

    def validate_other_parameters(self, parameters, valid_parameters, properties):
        for parameter, value in parameters.items():
            try:
                assert parameter in valid_parameters
            except AssertionError:
                message = "{} is not a valid parameter".format(parameter)
                raise errors.OctokitParameterError(message)
            if properties.get(parameter):
                self.validate_enum(parameter, value, properties)

    def validate_enum(self, parameter, value, properties):
        if properties.get(parameter).get("enum") and value not in properties.get(parameter).get("enum"):
            message = "{} is not a valid option for {}; must be one of {}".format(
                value, parameter, properties.get(parameter).get("enum")
            )
            raise errors.OctokitParameterError(message)

    def validate_schema(self, parameters, schema):
        if isinstance(parameters, list):
            self.validate_list(parameters, schema)
        if isinstance(parameters, dict):
            self.validate_dict(parameters, schema)
        self.validate_required_parameters(parameters, self.get_required_schema_properties(schema))
        self.type_match(parameters, schema.get("type"))

    def validate_dict(self, parameters, schema):
        for parameter, value in parameters.items():
            if isinstance(value, dict):
                self.validate_schema(value, schema.get("properties").get(parameter))
            if isinstance(value, list):
                self.validate_schema(value, schema.get("properties").get(parameter))

    def validate_list(self, parameters, schema):
        if not parameters and schema.get("items").get("required"):
            message = f"property is missing required items"
            raise errors.OctokitParameterError(message)
        for parameter in parameters:
            self.validate_schema(parameter, schema.get("items", schema))

    def type_match(self, symbol, expected_type):
        types = {"array": list, "object": dict, "string": str}
        if not isinstance(symbol, types.get(expected_type)):
            name = symbol.__class__.__name__
            message = f"{name} type does not match the schema type of {expected_type} for the data of {symbol}"
            raise errors.OctokitParameterError(message)

    def get_required_schema_properties(self, schema):
        if schema.get("type") == "object":
            return schema.get("required")

    def get_required_parameters(self, definition):
        return [p.get("name") for p in definition.get("parameters") if p.get("required") and p.get("in") in ["path"]]

    def validate_required_parameters(self, parameters, required_parameters):
        for required_parameter in required_parameters or []:
            try:
                assert required_parameter in parameters
            except AssertionError:
                self._raise_required_parameter(required_parameter)
            if parameters[required_parameter] is None:
                self._raise_must_have_value(required_parameter)

    def _raise_required_parameter(self, param):
        message = "{} is a required parameter".format(param)
        raise errors.OctokitParameterError(message)

    def _raise_must_have_value(self, required_parameter):
        message = "{} must have a value".format(required_parameter)
        raise errors.OctokitParameterError(message)

    def _form_url(self, values, _url, params):
        _values = dict(ChainMap(values, self._attribute_cache["url"]))
        filtered_kwargs = {k: v for k, v in _values.items() if params.get(k)}
        data_values = filtered_kwargs.copy()
        for name, value in filtered_kwargs.items():
            _url, subs = re.subn(fr"{{{name}}}", str(value), _url)
            if subs != 0:
                self._attribute_cache["url"][name] = data_values.pop(name)
        url = "{}{}".format(self.base_url, _url)
        return url, data_values

    def _get_data(self, kwargs, params):
        data = array_data = {}
        for parameter_name, parameter in params.items():
            if parameter.get("type") == "array":
                array_data = self._get_array_data(parameter_name, parameter, kwargs)
            data.update(self._get_default_data(parameter_name, parameter, kwargs))
        return dict(ChainMap(data, array_data, kwargs))

    def _get_default_data(self, parameter_name, parameter, kwargs):
        data = {}
        if parameter.get("in") in ["query", "body"] and not kwargs.get(parameter_name):
            if parameter.get("schema", parameter).get("default") is not None:
                data[parameter_name] = self._get_parameter_for_type(parameter.get("schema", parameter))
        return data

    def _get_array_data(self, parameter_name, parameter, kwargs):
        data = {}
        properties = parameter.get("items", parameter).get("properties")
        if properties:
            for name, property_data in properties.items():
                if property_data.get("default") and not kwargs.get(parameter_name):
                    data[parameter_name] = self._get_parameter_for_type(property_data)
        return data

    def _get_parameter_for_type(self, schema):
        if schema.get("type") == "boolean" and schema.get("default") == "true":
            return True
        if schema.get("type") == "boolean" and schema.get("default") == "false":
            return False
        return schema.get("default")

    def _data(self, data_kwargs, params, method):
        data = self._get_data(data_kwargs, params)
        if method == "get":
            return {"params": data}
        if method in ["post", "patch", "put", "delete"]:
            return {"data": json.dumps(data, sort_keys=True)}
        return {}

    def _setup_authentication(self, kwargs):
        authentication_schemes = {
            "basic": self._setup_basic_authentication,
            "token": self._setup_token_authentication,
            "installation": self._setup_installation_authentication,
            "app": self._setup_app_authentication,
        }
        if kwargs.get("auth"):
            authentication_schemes.get(kwargs.get("auth"))(kwargs)

    def _setup_basic_authentication(self, kwargs):
        assert kwargs["username"]
        assert kwargs["password"]
        self.username = kwargs["username"]
        self.password = kwargs["password"]
        self.auth = kwargs["auth"]

    def _setup_token_authentication(self, kwargs):
        assert kwargs["token"]
        self.token = kwargs["token"]
        self.auth = kwargs["auth"]

    def _setup_installation_authentication(self, kwargs):
        assert kwargs["app_id"]
        assert kwargs["private_key"]
        self.token, self.expires_at = self._app_auth_get_token(kwargs["app_id"], kwargs["private_key"])
        self.auth = kwargs["auth"]
        self.headers["accept"] = "application/vnd.github.machine-man-preview+json"

    def _setup_app_authentication(self, kwargs):
        assert kwargs["app_id"]
        assert kwargs["private_key"]
        self.jwt = self._app_auth_get_jwt(kwargs["app_id"], kwargs["private_key"])
        self.auth = kwargs["auth"]
        self.headers["accept"] = "application/vnd.github.machine-man-preview+json"

    def _app_auth_get_token(self, app_id, key):
        headers = {
            "Authorization": "Bearer {}".format(self._app_auth_get_jwt(app_id, key)),
            "Accept": "application/vnd.github.machine-man-preview+json",
        }
        installation_url = "{}/app/installations".format(self.base_url)
        installations = requests.get(installation_url, headers=headers).json()
        self.installation_id = [x.get("id") for x in installations if str(x.get("app_id")) == app_id].pop()
        installation_token_url = "{}/installations/{}/access_tokens".format(self.base_url, self.installation_id)
        response = requests.post(installation_token_url, headers=headers).json()
        return response["token"], response["expires_at"]

    def _app_auth_get_jwt(self, app_id, key):
        payload = {
            "iat": int(datetime.datetime.timestamp(datetime.datetime.now())),
            "exp": int(datetime.datetime.timestamp(datetime.datetime.now())) + (9 * 60),
            "iss": app_id,
        }
        return jwt.encode(payload, key, algorithm="RS256")

    def _auth(self, requests_kwargs):
        if getattr(self, "auth", None) == "basic":
            return {"auth": (self.username, self.password)}
        if getattr(self, "auth", None) in ["app", "token", "installation"]:
            _headers = {
                "app": {"Authorization": "Bearer {}".format(getattr(self, "jwt", None))},
                "token": {"Authorization": "token {}".format(getattr(self, "token", None))},
                "installation": {"Authorization": "token {}".format(getattr(self, "token", None))},
            }
            headers = requests_kwargs["headers"]
            headers.update(_headers.get(getattr(self, "auth", None)))
            return {"headers": headers}
        return {}

    def _get_parameters(self, definition, method):
        p = {}
        if definition.get("requestBody"):
            schema = definition.get("requestBody")["content"]["application/json"]["schema"]
            if schema["type"] == "object":
                body_parameters = {}
                for k, v in schema["properties"].items():
                    v["in"] = "body"
                    v["required"] = k in schema.get("required", {})
                    body_parameters.update({k: v})
                p.update(body_parameters)
        p.update(utils.parameter_transform(definition.get("parameters")))
        return p



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")).read()


setup(
    name="octokitpy",
    version="0.14.0",
    license="MIT license",
    description="Python client for GitHub API",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub("", read("README.rst")),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="Kyle Hornberg",
    author_email="kyle.hornberg@gmail.com",
    url="https://github.com/khornberg/octokit.py",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Utilities",
    ],
    keywords=["github", "octokit", "api"],
    install_requires=open("requirements.txt").readlines(),
    extras_require={},
)



import copy
import re
from collections import defaultdict

import requests
from octokit import utils
from octokit.base import Base
from octokit_routes import specifications

page_regex = re.compile(r'[\?\&]page=(\d+)[_&=%+\w\d]*>; rel="(\w+)"')


class Octokit(Base):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._create(specifications[kwargs.get("routes", "api.github.com")])
        self._setup_authentication(kwargs)

    def _create(self, definitions):
        classes = self._create_classes(definitions)
        bases = [object]
        for cls_name, class_attributes in classes.items():
            cls = type(cls_name, tuple(bases), class_attributes)
            setattr(self, cls_name, cls)

    def _create_classes(self, definitions):
        class_attributes = defaultdict(dict)
        for path, path_object in definitions["paths"].items():
            for method, method_object in path_object.items():
                cls_name, methods = self._get_class_methods(method_object, method, path)
                class_attributes[cls_name].update(methods)
        return class_attributes

    def _create_method(self, name, definition, method, path):
        def _api_call(*args, **kwargs):
            method_headers = kwargs.pop("headers") if kwargs.get("headers") else {}
            self.validate(kwargs, definition)
            requests_kwargs = {"headers": self._get_headers(method_headers)}
            parameter_map = self._get_parameters(definition, method)
            url, data_kwargs = self._form_url(kwargs, path, parameter_map)
            requests_kwargs.update(self._data(data_kwargs, parameter_map, method))
            requests_kwargs.update(self._auth(requests_kwargs))
            _response = getattr(requests, method)(url, **requests_kwargs)
            try:
                attributes = _response.json()
            except ValueError:
                attributes = _response.text
            new_self = copy.deepcopy(self)
            setattr(new_self, "_response", _response)
            setattr(new_self, "json", attributes)
            setattr(new_self, "response", new_self._convert_to_object(attributes))
            return new_self

        _api_call.__name__ = name
        _api_call.__doc__ = definition["description"]
        return _api_call

    def _get_names_from_operation_id(self, _object):
        _cls_name, _id_name = _object.get("operationId").split("/")
        cls_name = utils.snake_case(str(_cls_name))
        method_id_name = utils.snake_case(str(_id_name))
        return cls_name, method_id_name

    def _get_deprecated_methods(self, methods, method_object):
        deprecated_methods = {}
        if method_object.get("x-changes"):
            for change in method_object["x-changes"]:
                if change.get("type") == "operation":
                    before_cls, before_name = self._get_names_from_operation_id(change.get("before"))
                    after_cls, after_name = self._get_names_from_operation_id(change.get("after"))
                    if before_cls == after_cls and methods.get(after_name):
                        deprecated_methods.update({before_name: methods[after_name]})
        return deprecated_methods

    def _get_class_methods(self, method_object, method, path):
        cls_name, method_id_name = self._get_names_from_operation_id(method_object)
        method_name = utils.snake_case(str(method_object.get("summary")))
        methods = {
            method_id_name: self._create_method(method_id_name, method_object, method, path),
            method_name: self._create_method(method_name, method_object, method, path),
        }
        methods.update(self._get_deprecated_methods(methods, method_object))
        return cls_name, methods

    def _convert_to_object(self, item):
        if isinstance(item, dict):
            return type("ResponseData", (object,), {k: self._convert_to_object(v) for k, v in item.items()})
        if isinstance(item, list):
            return list((self._convert_to_object(value) for index, value in enumerate(item)))
        else:
            return item

    def set_pages(self, obj, previous_page_requested=None):
        response_headers = obj._response.headers
        links = response_headers.get("Link", None)
        if links:
            matches = re.findall(page_regex, links)
            if matches:
                for page, kind in matches:
                    setattr(obj, "{}_page".format(kind), int(page))
                setattr(obj, "pages", getattr(obj, "last_page", previous_page_requested))
                setattr(obj, "has_pages", True)
                setattr(obj, "current_page", previous_page_requested or getattr(obj, "next_page") - 1)
                setattr(obj, "is_last_page", obj.pages == obj.current_page)
            else:
                setattr(obj, "has_pages", False)
        return obj

    def paginate(self, obj, page=1, **kwargs):
        response = self.set_pages(obj(page=page, **kwargs))
        yield response.json
        if hasattr(response, "is_last_page"):
            while not response.is_last_page:
                response = self.set_pages(obj(page=response.next_page, **kwargs), response.next_page)
                yield response.json
